/*
 *  rgraph.c
 *  Anet
 *
 *  Created by Nicolas on Thu Aug 28 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 * $Id: rgraph.c,v 1.1 2003/09/10 09:20:16 nico Exp $
 * $Log: rgraph.c,v $
 * Revision 1.1  2003/09/10 09:20:16  nico
 * Renommage de mtrgraph en rgraph
 *
 * Revision 1.2  2003/09/10 09:12:57  nico
 * No log message.
 *
 * Revision 1.1.1.1  2003/09/09 14:41:04  nico
 * Version 0.9
 *
 * Revision 1.6  2003/09/09 07:18:05  nico
 * Version fonctionnelle, version de test
 *
 * Revision 1.5  2003/09/03 16:13:02  nico
 * Modification de la fonction de fusion de composantes pour elimination des liens depuis une composante vers la composante avec laquelle elle va fusionner
 *
 * Revision 1.4  2003/09/02 13:34:46  nico
 * No log message.
 *
 * Revision 1.3  2003/09/02 11:14:38  nico
 * Passage des donnees pscc vers RNET_PARAM
 *
 */

#include "rgraph.h"
#include "stack.h"

/* Macro for parameter*/
#define FIRSTBLOCK(rp) ((rp)->firstblock)
#define BLOCKCOUNT(rp) ((rp)->blockcount)
#define DFSNUM(rp, v) ((rp)->dfsnum[v])

#define IS_UNPROCESSED(rp,v) ((rp)->dfsnum[v] == 0xffffffffu)

#define ROOT		0x01
#define UNFINISHED	0x02

#define IS_UNFINISHED(rp, v) (rp->is[v] & UNFINISHED)
#define IS_ROOT(rp, v) (rp->is[v] & ROOT)

/* Data parameter for threaded functions */
typedef struct {
    ANET_RNET		*rnet;					/* used reduced net data structure */
    ANET_MASK		*subnet, *mask;			/* original subnet and isolated masked subnet blocks in MT mode */
    unsigned		compcount;				/* number of strong connected components */
    unsigned		firstblock, blockcount; /* first block and number of block involved, in MT mode */
    unsigned		*dfsnum, *pscc, count1; /* for DFS use (pscc is for englobing components) */
    stack			unfinished, root;		/* for DFS use */
    unsigned char	*is;					/* for DFS use : flags "is UNFINISHED" or "is ROOT" */
} RNET_PARAM;

void rnet_reduce_block(RNET_PARAM *rp);
void rnet_explore_block(RNET_PARAM *rp, unsigned v);
void rnet_collect_subnet_links(RNET_PARAM *rp);
void *rnet_process_blocks(RNET_PARAM *rp);
void rnet_flush_merged(RNET_PARAM *rp);
void *rnet_relabel_block(RNET_PARAM *rp);
void rnet_relabel_components(RNET_PARAM *rp);
void rnet_collect_inter_links(RNET_PARAM *rp);
void rnet_merge_blocks(RNET_PARAM *rp);
void rnet_explore_merged(RNET_PARAM *rp, unsigned c);

/* Initialize an allocated rnet data structure */
void rnet_init(ANET_RNET *rnet, ANET_CORE *core)
{
    rnet->compcount = 0;		/* no component */
    rnet->core = core;			/* base graph */
    rnet->size = AS_Size(core); /* size = base graph size */
	/* initialize components per block array to NULL */
    memset(rnet->compcounts, 0x00, ANET_MAX_CONCURRENCY * sizeof(unsigned));
	/* preallocate as many components as nodes in base graph (ie maximum component count) */
    rnet->scc = (RNET_SCC*)malloc(rnet->size * sizeof(RNET_SCC));
	/* initialize components to NULL */
    memset(rnet->scc, 0x00, rnet->size * sizeof(RNET_SCC));
	/* allocate and initialize to MAX_INT nodes component labels (0 is a valid index) */
    rnet->label = (unsigned*)malloc(rnet->size * sizeof(unsigned));
    memset(rnet->label, 0xff, rnet->size * sizeof(unsigned));
}

/* Release an allocated and build rnet */
void rnet_free(ANET_RNET *rnet)
{
    RNET_LINK   *link;  /* to go through predecessors list */
    unsigned    c;		/* component iterator */

    if (rnet == NULL) return;

    /* components have been relabeled */
    for (c = 0; c < rnet->compcount; c++)
        while (rnet->scc[c].links != NULL) {
            link = rnet->scc[c].links;
            rnet->scc[c].links = link->next;
            free(link);
        }
	/* release scc array and label list */
    if (rnet->scc) free(rnet->scc);
    if (rnet->label) free(rnet->label);
}

/* Return a preallocated and initialized rnet data structure address */
ANET_RNET *newRnet(ANET_CORE *core)
{
    ANET_RNET *rnet = (ANET_RNET*)malloc(sizeof(ANET_RNET));	/* allocate */
    
    rnet_init(rnet, core);  /* initialize */
    return rnet;
}

/* Multi-threaded recursive net reduction*/
void rnet_reduce(ANET_RNET *rnet, ANET_MASK *subnet)
{
    unsigned		*dfsnum, *pscc;		/* dfs numbers and parent scc indices arrays */
    unsigned char	*is;				/* is flags array */
    RNET_PARAM		rp;					/* rnet data parameter for MT mode */

    /* allocate dfsnum, pscc and is arrays */
    dfsnum = (unsigned*)malloc(AS_Size(CORE(rnet))*sizeof(unsigned));
    pscc = (unsigned*)malloc(AS_Size(CORE(rnet))*sizeof(unsigned));
    is = (unsigned char*)malloc(AS_Size(CORE(rnet))*sizeof(unsigned char));

	/* initialize parameter fields */
    rp.rnet = rnet; rp.subnet = rp.mask = subnet;
    rp.compcount = 0; rp.dfsnum = dfsnum; rp.is = is; rp.pscc = pscc;
	/* start from all blocks */
    rp.firstblock = 0; rp.blockcount = AS_BlockCount(CORE(rnet));

    if (AS_Concurrency(CORE(rnet)) > 1) {	/* if 2 or more threads */
		/* allocate memory for isolated subnet blocks */
        rp.mask = (ANET_MASK*)malloc(AS_Size(CORE(rnet))*sizeof(ANET_MASK));
    }

	/* process all blocks */
    rnet_process_blocks(&rp);

	/* if more than 1 block ... */
    if (AS_Concurrency(CORE(rnet)) > 1) {
		/* ... then relabel components for optimized access */
        rnet_relabel_components(&rp);
		/* and don't forget to release masked subnet blocks */
        free(rp.mask);
    } else {
		/* ... else memorize number of components */
        rnet->compcount = rp.compcount;
    }

	/* release potentially *huge* wasted memory in scc array */
    rnet->scc = (RNET_SCC*)realloc(rnet->scc, rnet->compcount*sizeof(RNET_SCC));
	/* free DFS data structures */
    free(dfsnum); free(pscc); free(is);
}

/* Process specified blocks */
void *rnet_process_blocks(RNET_PARAM *rp)
{
	/* if there's only one block to process, reduce it locally */
    if (rp->blockcount == 1) {
		/* buid corresponding block reduced net */
        rnet_reduce_block(rp);
    } else {	/* else split task in two */
        RNET_PARAM	rparam[2];
        pthread_t	task;
        pthread_attr_t	attr;
        unsigned	status;
        size_t		stack_size;

		/* same data parameter fields ... */
        rparam[1].rnet = rparam[0].rnet = rp->rnet;
        rparam[1].subnet = rparam[0].subnet = rp->subnet;
        rparam[1].mask = rparam[0].mask = rp->mask;
        rparam[1].dfsnum = rparam[0].dfsnum = rp->dfsnum;
        rparam[1].pscc = rparam[0].pscc = rp->pscc;
        rparam[1].is = rparam[0].is = rp->is;
		/* ...except that the block list is splitted in two */
        BLOCKCOUNT(rparam+1) = BLOCKCOUNT(rparam) = BLOCKCOUNT(rp)/2;
        FIRSTBLOCK(rparam) = FIRSTBLOCK(rp);
        FIRSTBLOCK(rparam+1) = FIRSTBLOCK(rparam) + BLOCKCOUNT(rparam);

		/* prepare a new thread for the second part of the list... */
        pthread_attr_init(&attr);
        pthread_attr_setstacksize(&attr, AS_BlockSize(CORE(rp->rnet),FIRSTBLOCK(rparam+1)) * 128);
        pthread_attr_getstacksize(&attr, &stack_size);
		/* and launch it */
        status = pthread_create(&task, &attr, (void*(*)(void*))(&rnet_process_blocks), rparam+1);
        if (status != 0)
            fprintf(stderr, "** ANET:CORE:MTRGRAPH> WARNING: Thread %d creation failed\n", FIRSTBLOCK(rparam+1));

		/* process first part of the list in the same time */
        rnet_process_blocks(rparam);
        
		/* when the two parts are processed */
        status = pthread_join(task, NULL);
		
		/* there is two locally build reduced net : we must merge cycles */
        if (status != 0)
            fprintf(stderr, "** ANET:CORE:MTRGRAPH> WARNING: Thread %d ending failed\n", FIRSTBLOCK(rparam+1));

        /* collect links between adjacent components */
        rnet_collect_inter_links(rp);
		/* merge blocks : merge cycles of the newly created components network */
        //printf("Merging blocks %d to %d:\n", FIRSTBLOCK(rp), FIRSTBLOCK(rp)+BLOCKCOUNT(rp)-1);
        rnet_merge_blocks(rp);
		/* flush links between merged components for faster list scrolling */
        rnet_flush_merged(rp);
    }
	/* here we've got the reduced net of the rp's block list */
    return NULL;
}

/* Remove obsolet predecessor links (after component merging) */
void rnet_flush_merged(RNET_PARAM *rp)
{
    unsigned    b, c;
    ANET_RNET   *rnet = rp->rnet;
    RNET_LINK   *link, *tmplink;

	/* for each block b of rp block list */
    for (b = FIRSTBLOCK(rp); b < FIRSTBLOCK(rp)+BLOCKCOUNT(rp); b++) {
		/* for each component of block b */
        for (c = AS_BlockStart(CORE(rnet), b); c < AS_BlockStart(CORE(rnet), b)+ rnet->compcounts[b]; c++) {
			/* if component c is an englobing component and c has predecessors... */
            if (rp->pscc[c] == c && rnet->scc[c].links != NULL) {
				/* remove links from englobed component and from self! */
                link = rnet->scc[c].links;
                while (link != NULL && (link->scc != rp->pscc[link->scc] || link->scc == c)) {
                    tmplink = link;
                    link = link->next;
                    free(tmplink);
                }
                rnet->scc[c].links = link;
                if (link) while (link->next != NULL) {
                    if (link->next->scc != rp->pscc[link->next->scc] || link->next->scc == c) {
                        tmplink = link->next;
                        link->next = tmplink->next;
                        free(tmplink);
                    }
                    link = link->next;
                }
            }
        }
    }
}

/* Relabel block after global component relabelling (for MT mode) */
void *rnet_relabel_block(RNET_PARAM *rp)
{
    ANET_RNET   *rnet = rp->rnet;
    register unsigned begin, end, i, *label = rp->rnet->label, *dfsnum = rp->dfsnum;

    begin = AS_BlockStart(CORE(rnet), FIRSTBLOCK(rp));
    end = AS_BlockStart(CORE(rnet), FIRSTBLOCK(rp) + 1);

	/* reusing dfsnum to spare memory consumption (see rnet_relabel components)*/
    for (i = begin; i < end; i++)
        if (label[i] < 0xffffffffu) label[i] = dfsnum[label[i]];

    return NULL;
}

/* Re-indexation of components for optimized access */
void rnet_relabel_components(RNET_PARAM *rp)
{
    RNET_PARAM  rpl[ANET_MAX_CONCURRENCY];
    pthread_t   task[ANET_MAX_CONCURRENCY];
    ANET_RNET   *rnet = rp->rnet;
    unsigned    b, c, status;
    
	/* for each block b */
    for (b = 0; b < rp->blockcount; b++) {
		/* for each component c of block b */
        for (c = AS_BlockStart(rnet->core, b); c < AS_BlockStart(rnet->core, b)+rnet->compcounts[b]; c++) {
			/* if c is an englobing component */
            if (rp->pscc[c] == c) {
				/* add it to the final list of components (scc array) */
                rnet->scc[rnet->compcount] = rnet->scc[c];
				/* put its new index (rnet->compcount) in dfsnum[c] (for use in rnet_relabel_block function) */
                rp->dfsnum[c] = rnet->compcount++;
            } else {
				/* put its englobing component index in dfsnum[c] (for use in rnet_relabel_block function) */
                rp->dfsnum[c] = rp->dfsnum[rp->pscc[c]];
            }
        }
		/* lets modify labels in rnet to reflect this re-indexation in block b */
        rpl[b].rnet = rnet; rpl[b].dfsnum = rp->dfsnum; rpl[b].firstblock = b; rpl[b].blockcount = 1;
		/* it can be done asynchronously (in the same time other blocks are processed) */
        status = pthread_create(task+b, NULL, (void*(*)(void*))(&rnet_relabel_block), rpl+b);
        if (status != 0)
            fprintf(stderr, "** ANET:CORE:MTRGRAPH> WARNING: Thread %d creation failed\n", b);
    }
    
    for (b = 0; b < rp->blockcount; b++) {
        status = pthread_join(task[b],NULL);
        if (status != 0)
            fprintf(stderr, "** ANET:CORE:MTRGRAPH> WARNING: Thread %d ending failed\n", b);
    }
}

/* Local graph reduction : build a reduced net on an isolated block */
void rnet_reduce_block(RNET_PARAM *rp)
{
    unsigned	begin, end, v;

	/* get first and last nodes indices */
    begin = AS_BlockStart(CORE(rp->rnet), FIRSTBLOCK(rp));
    end = AS_BlockStart(CORE(rp->rnet), FIRSTBLOCK(rp) + BLOCKCOUNT(rp));

    if (rp->subnet != rp->mask) { /* that means there are at least two blocks */
		/* build block corresponding subnet by desactivating arcs (called "interblock-arcs") with adjacent blocks */
		/* it can be done at a block level to leverage more parallelism */
        memcpy(rp->mask + begin, rp->subnet + begin, AS_BlockSize(CORE(rp->rnet),FIRSTBLOCK(rp)) * sizeof(ANET_MASK));
        /* mask interblock-arcs (arcs between 2 different blocks) */
        for (v = begin; v < begin + AS_Width(CORE(rp->rnet)); v++)
            rp->mask[v] &= ~(mNW|mN|mNE);
        for (v = end-1; v >= end - AS_Width(CORE(rp->rnet)); v--)
            rp->mask[v] &= ~(mSW|mS|mSE);
    }

	/* allocate and initialize stacks for dfs (tarjan algorithm) */
    stack_alloc(rp->unfinished, end-begin+1);
    stack_init(rp->unfinished);
    stack_alloc(rp->root, end-begin+1);
    stack_init(rp->root);
	/* initialize is and dfsnum values (at a block level for more parallelism) */
    memset(rp->is + begin, 0x00, (end-begin)*sizeof(unsigned char));
    memset(rp->dfsnum + begin, 0xff, (end-begin)*sizeof(unsigned));
    rp->count1 = 0;
    rp->compcount = begin;

	/* explore unprocessed nodes of the block */
    for (v = begin; v < end; v++) {
        if (rp->subnet[v] != 0 && IS_UNPROCESSED(rp,v))
            rnet_explore_block(rp, v);
    }

	/* count number of components for that block */
    rp->rnet->compcounts[FIRSTBLOCK(rp)] = rp->compcount - begin;

	/* collect inter-components links from nodes arcs */
    rnet_collect_subnet_links(rp);

	/* release stacks used memory */
    stack_free(rp->unfinished);
    stack_free(rp->root);
    //printf("Block %d contains %d components\n", FIRSTBLOCK(rp), rp->rnet->compcounts[FIRSTBLOCK(rp)]);
}

/* DFS node exploration for scc detection (Tarjan) */
void rnet_explore_block(RNET_PARAM *rp, unsigned v)
{
    int 	*neighbor;
    unsigned	n, w;

	/* attributes v a dfs number */
    DFSNUM(rp, v) = ++(rp->count1);
	/* in the beginning v is unfinished, and a component root */
    stack_push(rp->unfinished, v);
    rp->is[v] |= UNFINISHED;
    stack_push(rp->root, v);
    rp->is[v] |= ROOT;

	/* get v neighbors list */
    neighbor = t_neighbors_rel(v);
	/* for each nth neighbor */
    for (n = 0; n < ANET_MASK_SIZE; n++) {
		/* if its link with v is activated */
        if (rp->mask[v] & (1 << n)) {
			/* get its abslute indice w */
            w = v+neighbor[n];
			/* if w is also unprocessed resume exploration through it */
            if (IS_UNPROCESSED(rp, w)) {
                rnet_explore_block(rp,w);
            } else { /* we've already seen w */
				/* if it's still active */
                if (IS_UNFINISHED(rp, w)) {
					/* all nodes explores from w belong to the same component ! */
                    while (DFSNUM(rp, stack_head(rp->root)) > DFSNUM(rp, w)) {
						/* they can't be root (extract them from the root stack) */
                        rp->is[stack_pop(rp->root)] &= ~ROOT;
					}
                }
            }
        }
    }

	/* all that could be explored from v is explored and processed */
	/* if v is still in the root stack... */
    if (v == stack_head(rp->root)) {
		/* ... then that's because it's the root of a new strong connected component ! */
		/* but not anymore !*/
        rp->is[stack_pop(rp->root)] &= ~ROOT;
		/* let's pop unfinished nodes w until we reach down to v */
        do {
            w = stack_pop(rp->unfinished);
			/* now w is finished... */
            rp->is[w] &= ~UNFINISHED;
			/* ...and it belongs to the component rp->compcount*/
            rp->rnet->label[w] = rp->compcount;
        } while (w != v);
		/* for now the component is its own englobing component */
        rp->pscc[rp->compcount] = rp->compcount;
		/* ok, don't forget we've got one more component ! */
        rp->compcount++;
    }
}

/* Macro definition for inserting component links with duplicate removal */
#define COLLECT_SUBNET_LINK(cl, cd) { \
    tmp = scc[cl].links; \
    if (tmp == NULL) { \
        tmp = scc[cl].links = (RNET_LINK*)malloc(sizeof(RNET_LINK)); \
        tmp->scc = cd; \
        tmp->next = NULL; \
    } else { \
        while (tmp->scc != cd && tmp->next != NULL) { tmp = tmp->next; } \
        if (tmp->scc != cd) { \
            tmp = scc[cl].links; \
            scc[cl].links = (RNET_LINK*)malloc(sizeof(RNET_LINK)); \
            (scc[cl].links)->scc = cd; \
            (scc[cl].links)->next = tmp; \
        } \
    } \
} \

/* Collect links between adjacent notes from different components */
/* Then buid list of components predecessor links */
void rnet_collect_subnet_links(RNET_PARAM *rp)
{
    unsigned	begin, end, v, n;
    int			*neighbor;
    unsigned	*label = rp->rnet->label;
    RNET_SCC	*scc = rp->rnet->scc;
    RNET_LINK	*tmp;
    ANET_MASK	*mask = rp->mask;

	/* get first and last node to check */
    begin = AS_BlockStart(CORE(rp->rnet), FIRSTBLOCK(rp));
    end = AS_BlockStart(CORE(rp->rnet), FIRSTBLOCK(rp) + BLOCKCOUNT(rp));

	/* for each node of the block */
    for (v = begin; v < end; v++) {
		/* if it has incident arcs (ie predecessors in the subnet) */
        if (mask[v]) {
			/* get v neighbors list */
            neighbor = t_neighbors_rel(v);
			/* for each nth neighbor */
            for (n = 0; n < ANET_MASK_SIZE; n++) {
				/* if incident arc is activated and neighbor belongs to a component that is different from v component */
                if ((mask[v] & (1 << n)) && (label[v+neighbor[n]] < 0xffffffffu) && (label[v] != label[v+neighbor[n]])) {
					/* add that link between the two component without duplication */
                    COLLECT_SUBNET_LINK(label[v], label[v+neighbor[n]])
				}
            }
        }
    }
}

/* Get englobing component of component c after several merges */
inline unsigned rnet_pscc(unsigned *pscc, unsigned c)
{
	/* while c is not an englobing component */
    while (c != pscc[c])
		c = pscc[c];
    return c;
}

/* Collect links between components from 2 different adjacent blocks before possible merge */
/* First step of local rnet merge */
void rnet_collect_inter_links(RNET_PARAM *rp)
{
    unsigned	begin, end;
    int			*neighbor;
    unsigned	*label = rp->rnet->label;
    RNET_SCC	*scc = rp->rnet->scc;
    ANET_RNET	*rnet = rp->rnet;
    RNET_LINK	*tmp;
    ANET_MASK	*subnet = rp->subnet;
    register unsigned	v, vscc;

	/* get the first and last nodes of first block (north) linked with noes of the second block */
    end = AS_BlockStart(CORE(rnet), FIRSTBLOCK(rp)+BLOCKCOUNT(rp)/2);
    begin = end - AS_Width(CORE(rnet));

	/* for each node */
    for (v = begin; v < end; v++) {
        vscc = label[v];
		/* if it belongs to a component (vscc)  */
        if (vscc < 0xffffffffu) {
			/* get the neighbors list */
            neighbor = t_neighbors_rel(v);
			/* get vscc's englobing component */
            vscc = rnet_pscc(rp->pscc, vscc);
			/* if node v's links with neighbors from other blocks are activated connect the components */
            if (subnet[v] & (1 << SW))
                COLLECT_SUBNET_LINK(vscc, rnet_pscc(rp->pscc,label[v+neighbor[SW]]))
            if (subnet[v] & (1 << S))
                COLLECT_SUBNET_LINK(vscc, rnet_pscc(rp->pscc,label[v+neighbor[S]]))
            if (subnet[v] & (1 << SE))
                COLLECT_SUBNET_LINK(vscc, rnet_pscc(rp->pscc,label[v+neighbor[SE]]))
        }
    }

	/* do the same with the second block (south) */
    begin = end;
    end = begin + AS_Width(CORE(rnet));
    
    for (v = begin; v < end; v++) {
        vscc = label[v];
        if (vscc < 0xffffffffu) {
            neighbor = t_neighbors_rel(v);
            vscc = rnet_pscc(rp->pscc, vscc);
            if (subnet[v] & (1 << NW))
                COLLECT_SUBNET_LINK(vscc, rnet_pscc(rp->pscc,label[v+neighbor[NW]]))
            if (subnet[v] & (1 << N))
                COLLECT_SUBNET_LINK(vscc, rnet_pscc(rp->pscc,label[v+neighbor[N]]))
            if (subnet[v] & (1 << NE))
                COLLECT_SUBNET_LINK(vscc, rnet_pscc(rp->pscc,label[v+neighbor[NE]]))
        }
    }
}

/* Merge two adjacent blocks : merge cycles of two connected local rnets */
/* Second step of local rnet merge : use Tarjan algorithm again ! */
void rnet_merge_blocks(RNET_PARAM *rp)
{
    unsigned	begin, end, b, c;
    ANET_RNET	*rnet = rp->rnet;

	/* get first and last nodes of merged block */
    begin = AS_BlockStart(CORE(rnet), FIRSTBLOCK(rp));
    end = AS_BlockStart(CORE(rnet), FIRSTBLOCK(rp)+BLOCKCOUNT(rp));

	/* allocate and initialize stacks for dfs algorithm */
    stack_alloc(rp->unfinished, end-begin+1);
    stack_init(rp->unfinished);
    stack_alloc(rp->root, end-begin+1);
    stack_init(rp->root);
	/* initialize is and dfsnum arrays */
    memset(rp->is + begin, 0x00, (end-begin)*sizeof(unsigned char));
    memset(rp->dfsnum + begin, 0xff, (end-begin)*sizeof(unsigned));
    rp->count1 = 0;
    rp->compcount = begin; //not used here
//printf("Begin merged exploration\n");
	/* for each elementary block b composing the merged block */
    for (b = FIRSTBLOCK(rp); b < FIRSTBLOCK(rp)+BLOCKCOUNT(rp); b++) {
//        printf("  Block %d :\n", b);
		/* for each component c of that block b */
        for (c = AS_BlockStart(CORE(rnet), b); c < AS_BlockStart(CORE(rnet), b)+ rnet->compcounts[b]; c++) {
//            printf("    SCC %d :\n", c);
			/* if c is an englobing component and c is not already explored */
            if (rp->pscc[c] == c && rp->dfsnum[c] == 0xffffffffu) {
//                printf("      exploring...\n");
				/* start cycle exploring */
                rnet_explore_merged(rp, c);
            }
        }
    }
//printf("End merged exploration\n");
    
    stack_free(rp->unfinished);
    stack_free(rp->root);
}

/* DFS component exploration for cycle detection (Tarjan) */
/* Exactly the same function as rnet_explore_block, but apllied to crnet components ! */
void rnet_explore_merged(RNET_PARAM *rp, unsigned c)
{
    unsigned	parent;
    RNET_LINK	*link;
    ANET_RNET	*rnet = rp->rnet;

    DFSNUM(rp, c) = ++(rp->count1);
    stack_push(rp->unfinished, c);
    rp->is[c] |= UNFINISHED;
    stack_push(rp->root, c);
    rp->is[c] |= ROOT;

    link = rnet->scc[c].links;
    while (link != NULL) {
        parent = link->scc;
//        printf("        parent %d : ", parent);
        if (rp->pscc[parent] == parent && rp->dfsnum[parent] == 0xffffffffu) { //not explored
//            printf("exploring\n");
            rnet_explore_merged(rp, parent);
        } else {
//            printf("active\n");
            if (IS_UNFINISHED(rp, parent)) { //active
                while (DFSNUM(rp, stack_head(rp->root)) > DFSNUM(rp, parent))
                        rp->is[stack_pop(rp->root)] &= ~ROOT;
            }
        }
        link = link->next;
    }
    
    if (c == stack_head(rp->root)) { /* we have detected a strong connected component of components ! */
//        printf("New scc %d\n", c);
        rp->is[stack_pop(rp->root)] &= ~ROOT;
        parent = stack_pop(rp->unfinished);
        while (c != parent) {
            rp->is[parent] &= ~UNFINISHED;
            //printf("La composante %d absorbe la composante %d\n", c, parent);
			/* component c now englobes the parent component */
            rp->pscc[parent] = c;
            /* merge lists of incident links */
            link = rnet->scc[c].links;
            if (link) {
                while (link->next != NULL) link = link->next;
                link->next = rnet->scc[parent].links;
            } else {
                rnet->scc[c].links = rnet->scc[parent].links;
            }
            parent = stack_pop(rp->unfinished);
        }
        rnet->scc[parent].links = NULL;
        rp->is[parent] &= ~UNFINISHED;
        rp->compcount++; //not used here
    }
}

void rnet_print(ANET_RNET *rnet, unsigned width, unsigned flag)
{
    unsigned i;
    
    if (rnet == NULL) return;
    printf("rnet : %d components\n", rnet->compcount);
    if (flag) {
        for (i = 0; i < rnet->size; i++) {
            if (i % width) printf("%3d ", rnet->label[i]); else printf("\n%3d ", rnet->label[i]);
        }
        printf("\n");
    }
}
