/*
 *  vcontext.c
 *  Anet
 *
 *  Created by Nicolas on Tue Apr 20 2004.
 *  Copyright (c) 2004 IEF. All rights reserved.
 *
 */

#include "vcontext.h"
#include "vgraph.h"

VNET_CONTEXT_STACK *newVContextStack(struct ANET_VNET *vnet)
{
	VNET_CONTEXT_STACK *stack = (VNET_CONTEXT_STACK*)malloc(sizeof(VNET_CONTEXT_STACK));
	
	vc_init(stack, vnet);
	return stack;
}

/* allocate memory and initialize context object */
void vc_init(VNET_CONTEXT_STACK *ctxs, struct ANET_VNET *vnet)
{
	if (ctxs == NULL) return;
	
	ctxs->depth = 0;
	ctxs->stack = (VNET_CONTEXT*)malloc(ANET_MAX_CTXDEPTH*sizeof(VNET_CONTEXT));
	ctxs->current = ctxs->stack;
	ctxs->size = vnet->compcount;
	ctxs->core = vnet;
	vc_setEverywhere(ctxs->current, ctxs->size);
}

/* free context object allocated memory */
void vc_free(VNET_CONTEXT_STACK *ctxs)
{
	if (ctxs == NULL) return;
	
	/* Warn if we detect unbalanced where/endwhere pairs */
	if (ctxs->depth != 0) {
		fprintf(stderr, "** VNET:CORE:CONTEXT> Warning: Unbalanced WHERE/ENDWHERE pairs.\n" );
	}
	/* Free memory space allocated for execution context */
	if (ctxs->stack != NULL) { free(ctxs->stack); ctxs->stack = NULL; }
	ctxs->current = NULL;
	ctxs->depth = 0;
}

/* set context object to EVERYWHERE */
void vc_setEverywhere(VNET_CONTEXT *ctx, unsigned size)
{
	ctx->type = ctxtEverywhere;
	ctx->cond = NULL;
	ctx->size = size;
}

/* set context object to NOWHERE */
void vc_setNowhere(VNET_CONTEXT *ctx)
{
	ctx->type = ctxtNowhere;
	ctx->cond = NULL;
	ctx->size = 0;
}

/* pop last context from the context stack */
void vc_PopContext(VNET_CONTEXT_STACK *ctxs)
{
	VNET_CONTEXT *current = vc_getCurrentContext(ctxs);
	if(vc_isPvarContext(current)) {
		free(current->cond);
		current->cond = NULL;
		current->type = ctxtNowhere;
		current->size = 0;
	}
	ctxs->depth-- ;
	ctxs->current--;
}

typedef struct {
    unsigned char	*cond;
    //ANET_CORE		*core;
    unsigned char	*current, *previous;
    unsigned		size, flag;
} CONTEXT_PARAM;


/* put a new context on top of the stack from boolean values */
void vc_Where(VNET_CONTEXT_STACK *ctxs, unsigned char *cond)
{
	unsigned i;
	VNET_CONTEXT *current, *previous;
	
	// get adress of current context that will become the previous one
	previous = vc_getCurrentContext(ctxs);
	
	vc_PushContext(ctxs);
	current = vc_getCurrentContext(ctxs);  // get new current context
	
	/* if previous context was "nowhere" then this is "nowhere" too ! */
	if (vc_isNowhere(previous)){
		vc_setNowhere(current);
		return;
	}

	current->cond = (unsigned char*)malloc(ctxs->size*sizeof(unsigned char));
			
	// if previous context was from a boolean pvar then proceed to a
	// logical AND between "cond" and it
	if (vc_isPvarContext(previous)) {
		for (i = 0; i < ctxs->size; i++)
			current->cond[i] = (unsigned char)((int)previous->cond[i] && (int)cond[i]);
	} else { /* scan all "cond" to check active vertices */
		memcpy(current->cond, cond, ctxs->size*sizeof(unsigned char));
	}
	
	current->size = ctxs->size;
	current->type = ctxtPvarBool;
	if (current->size > 0) current->type = ctxtPvarBool; else vc_setNowhere(current);
}

/* pop last context and make previous the current context */
void vc_Endwhere(VNET_CONTEXT_STACK *ctxs) 
{
	vc_PopContext(ctxs) ;
}


/* replace current context by its complement -> ELSEWHERE */
void vc_Elsewhere(VNET_CONTEXT_STACK *ctxs)
{
	unsigned i;
	VNET_CONTEXT *current, *previous = NULL;
	
	if (ctxs->depth == 0 || ctxs->depth > ANET_MAX_CTXDEPTH) {
		fprintf(stderr, "ANET> Error : No Where before Elsewhere or ANET_MAX_CTXDEPTH overflowed\n");
	}
	
	// get current and previous contexts
	current = vc_getCurrentContext(ctxs);
	previous = vc_getPreviousContext(ctxs);
	
	// if current context is EVERYWHERE
	if (vc_isEverywhere(current)){
		fprintf(stderr,"ANET> Error : Elsewhere in everywhere context !\n");
		return;
	}
	
	// if current context is NOWHERE
	if (vc_isNowhere(current)){
		if (vc_isNowhere(previous)) return; // if previous context is nowhere, no change 
		if (vc_isEverywhere(previous)) // if previous context is everywhere
			vc_setEverywhere(current, ctxs->size);
		else{ // previous context is "somewhere"
			  // then current context becomes previous context copy
			current->cond = (unsigned char*)malloc(sizeof(unsigned char)*ctxs->size);
			memcpy(current->cond, previous->cond, sizeof(unsigned char)*ctxs->size);
			current->type = previous->type;
			current->size = previous->size;
		}
		return;
	}
	

	// current context is SOMEWHERE
	// if previous context is EVERYWHERE
	if(vc_isEverywhere(previous)){ /* complement current context */
		for (i = 0; i < ctxs->size; i++)
			current->cond[i] = (unsigned char)(!(current->cond[i]));
	}else{ /* combine previous and current contexts */
		for (i = 0; i < ctxs->size; i++)
			current->cond[i] = (unsigned char)(!(current->cond[i]) && (int)(previous->cond[i]));
	}
}
